#!/usr/bin/env python3

from string import printable as ascii_printable

def _collapse_runs(nums):
    nums = sorted(nums)
    ret = []
    while nums:
        start = n = nums[0]
        nums = nums[1:]
        while nums and n+1 == nums[0]:
            n = nums[0]
            nums = nums[1:]
        if start == n: ret.append(n)
        else: ret.append((start,n))
    return ret
assert _collapse_runs([3,4,5,9,10,11,14]) == [(3,5),(9,11),14]

def _fmt_group(chars):
    if len(chars) == 0: raise
    if len(chars) == 1: return chars[0]
    if any(char not in ascii_printable for char in chars): raise
    digits, uppers, lowers, etcs = [], [], [], []
    for char in chars:
        if   char.isdigit(): digits.append(char)
        elif char.isupper(): uppers.append(char)
        elif char.islower(): lowers.append(char)
        else: etcs.append(char)
    ret = ''.join('{x[0]}-{x[1]}'.format(x=x) if isinstance(x,tuple) else str(x) for x in _collapse_runs(map(int,digits)))
    ret += ''.join('{chr(x[0])}-{chr(x[1])}'.format(x=x) if isinstance(x,tuple) else chr(x) for x in _collapse_runs(map(ord,lowers)))
    ret += ''.join('{chr(x[0])}-{chr(x[1])}'.format(x=x) if isinstance(x,tuple) else chr(x) for x in _collapse_runs(map(ord,uppers)))
    ret += ''.join(sorted(etcs))
    return '[' + ret +']'

def get_globs(data):
    data = list(set(data))
    lens = {}
    for s in data: lens.setdefault(len(s), []).append(s)
    for length,strings in sorted(lens.items()):

        things = [list(s) for s in strings]
        for i in reversed(range(length)):
            things = [(tuple(thing[:i]), thing[i], tuple(thing[i+1:])) for thing in things]
            groups = {}
            for prefix, char, suffix in things:
                groups.setdefault((prefix,suffix), []).append(char)
            things = [list(prefix) + [tuple(sorted(chars))] + list(suffix) for (prefix,suffix),chars in groups.items()]

        for thing in sorted(things):
            yield ''.join(_fmt_group(g) for g in thing)


if __name__ == '__main__':

    # TODO: make expandglob(), and test that for arbitrary lists of strings, expandglob(unglob(strings)) == strings
    # TODO: run tests with pytest

    datasets = [
            ['{code:03}'.format(code=code) for code in range(0,100+1)],

            ['{code}'.format(code=code) for code in range(0,100+1)],

            (
                '310 311 312 313 314 315 316 317 318 319 320 321 322 323 324 325 326 327 328 329 330 331 332 333 334 335 336 337 338 339 340 341 342 343 344 345 346 347 348 349 350 351 352 353 354 355 356 357 358 359 360 361 362 363 364 365 366 367 368 369 370 371 372 373 374 375 376 377 378 379 380 381 382 383 384 385 386'.split() +
                '308/200 309/200 310/200 311/200 312/200 313/200 314/200 315/200 316/200 317/200 318/200 319/200 320/200 321/200 322/200 323/200 324/200 325/200 326/200 327/200 328/200 329/200 330/200 331/200 332/200 333/200 334/200 335/200 336/200 337/200 338/200 339/200 340/200 341/200 342/200 343/200 344/200 345/200 346/200 347/200 348/200 349/200 350/200 351/200 352/200 353/200 354/200 355/200 356/200 357/200 358/200 359/200 360/200 361/200 362/200 363/200 364/200 365/200 366/200 367/200 368/200 369/200 370/200 371/200 372/200 373/200 374/200 375/200 376/200 377/200 378/200 379/200 380/200 381/200 382/200 383/200 384/200 385/200 386/200 387/200 388/200 389/200'.split()
            )
    ]
    for data in datasets:
        for glob in get_globs(data):
            print(glob)
        print()




def iterlen(iterator):
    '''returns length of iterator'''
    return sum(1 for _ in iterator)

def get_extent(iterator):
    '''returns (min, max) of iterator'''
    iterator = iter(iterator)
    m = M = next(iterator)
    for elem in iterator:
        if elem < m: m = elem
        elif elem > M: M = elem
    return (m,M)




import shelve, json, os.path


def _jsonify(obj):
    return json.dumps(obj, default=_jsonify_default)
def _jsonify_default(obj):
    try:
        import pandas as pd
    except ImportError:
        pass
    else:
        if isinstance(obj, pd.DataFrame):
            return {'type=pandas.DataFrame': obj.to_json()}
    raise TypeError('Object of type {obj.__class__.__name__} is not JSON serializable!'.format(**locals()))


def shelve_cache(func):
    def new_f(*args, **kwargs):
        key = _jsonify([args, kwargs])
        with shelve.open('.cache-{func.__name__}.shelve'.format(**locals())) as shelf:
            if key not in shelf: shelf[key] = func(*args, **kwargs)
            return shelf[key]
    return new_f


def cached_generator(record_maker=lambda x:x):
    '''
    caches a generator into jsonlines format.
    calls record_maker on each un-serialized line before yielding it.
    doesn't track args, so don't use any.
    '''
    def decorator(func):
        cache_fpath = os.path.join('.cache-{func.__name__}.jsonlines'.format(**locals()))
        def newf(*args, **kwargs):
            if args or kwargs: raise Exception('cached_generator() cannot handle args or kwargs')
            if not os.path.exists(cache_fpath):
                with open(cache_fpath+'.tmp', 'wt') as f:
                    for r in func():
                        f.write(json.dumps(r) + '\n')
                        yield r
                os.rename(cache_fpath+'.tmp', cache_fpath)
            else:
                with open(cache_fpath, 'rt') as f:
                    for line in f:
                        yield record_maker(json.loads(line))
        return newf
    return decorator




class lru_cache_set:
    '''
    A set that is limited to size `maxsize` by removing the least-recently-accessed item when necessary.
    `.add()`ing an item that is already present simply updates the access time for that item.
    Checking `item in cache` updates the access time for that item.

    Items must be `hash()`able.

    It is implemented with a dict and a circular-doubly-linked-list of items, sorted by how recently they were accessed.
    '''
    OLDER, NEWER, KEY = 0, 1, 2

    def __init__(self, maxsize):
        self.maxsize = maxsize; assert maxsize >= 2
        self._cache = {} # like `{<key> -> <item>}`
        # <item> is like [<older>, <newer>, <key>] (where older & newer are like <item>)
        # self._root is a special <item>:
        # self._root[OLDER] is the newest item, and self._root[NEWER] is the oldest.
        self._root = []; self._root[:] = [self._root, self._root, None]

    def __repr__(self):
        keys = []
        item = self._root[self.NEWER]
        while item is not self._root:
            assert item[self.NEWER][self.OLDER] is item
            keys.append(item[self.KEY])
            item = item[self.NEWER]
        return repr(keys)

    def __contains__(self, key):
        # checks whether we contain a key and also makes it the newest
        item = self._cache.get(key, None)
        if item is None:
            return False
        self._move_item_to_newest(item)
        return True

    def add(self, key):
        # add a key.  if we already have it, just make it newest.
        NEWER, OLDER, KEY = self.NEWER, self.OLDER, self.KEY
        if key in self._cache:
            self._move_item_to_newest(self._cache[key])

        else:
            if len(self._cache) == self.maxsize:
                # re-use item that used to be oldest
                item = self._root[NEWER]
                del self._cache[item[KEY]]
                # splice it out
                item[OLDER][NEWER] = item[NEWER]
                item[NEWER][OLDER] = item[OLDER]
                item[:] = [self._root[OLDER], self._root, key]
            else:
                item = [self._root[OLDER], self._root, key]
            self._cache[key] = item
            self._root[OLDER][NEWER] = item
            self._root[OLDER] = item

    def _move_item_to_newest(self, item):
        OLDER, NEWER = self.OLDER, self.NEWER
        if item is self._root[OLDER]: return
        # splice out item
        item[OLDER][NEWER] = item[NEWER]
        item[NEWER][OLDER] = item[OLDER]
        # splice in item as newest
        item[OLDER] = self._root[OLDER]
        item[NEWER] = self._root
        self._root[OLDER][NEWER] = item
        self._root[OLDER] = item
        self._root[OLDER] = item


class _lru_cache_set_slow:
    'a simple but inefficient implementation of lru_cache_set'
    def __init__(self, maxsize):
        self.maxsize = maxsize
        self._keys = [] # oldest item is at self._keys[0], newest is at self._keys[-1]
    def __repr__(self):
        return repr(self._keys)
    def add(self, key):
        if key in self._keys:
            self._keys.remove(key)
        self._keys.append(key)
        if len(self._keys) > self.maxsize:
            self._keys.pop(0)
    def __contains__(self, key):
        if key not in self._keys:
            return False
        self._keys.remove(key)
        self._keys.append(key)
        return True


if __name__ == '__main__':

    c1 = lru_cache_set(4)
    c2 = _lru_cache_set_slow(4)
    assert repr(c1) == repr(c2)
    for key in 'b b a b c d e b d b f a b c d'.split():
        c1.add(key)
        c2.add(key)
        print(key, c1)
        assert repr(c1) == repr(c2)

    for key in 'abcdefg':
        has1 = key in c1
        has2 = key in c2
        assert has1 == has2, (key, has1, has2, c1, c2)
        assert repr(c1) == repr(c2)
        print('{}{}{}'.format('-+'[int(has1)], key, c1))




import boltons.funcutils

def assign(func):
    return func()
def assign_list(func):
    return list(func())

def only_once(func):
    already_ran = [False] # TODO: use `nonlocal` or `global` or whatever to avoid `[0]`
    @boltons.funcutils.wraps(func)
    def f(*args, **kwargs):
        if not already_ran[0]:
            already_ran[0] = True
            return func(*args, **kwargs)
    return f

def list_from_iter(func):
    @boltons.funcutils.wraps(func)
    def f(*args, **kwargs):
        return list(func(*args, **kwargs))
    return f
